# IRAF dummy standard system package script task declarations.
#
# This file is used when no IRAF is available.
#
task  language.pkg = "language$language.cl"
task  system.pkg   = "system$system.cl"
system
keep
